local shader = {}
  
  shader.energy = love.graphics.newShader([[
    
    extern float power;
    
    vec4 effect( vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords )
    {
      vec4 pixel = Texel(texture, texture_coords ); //This is the current pixel color
      number average = (pixel.r*color.r+pixel.g*color.g+pixel.b*color.b)/3.0;
      pixel.r = mix(pixel.r, average*1.6, power);
      pixel.g = mix(pixel.g, average*1.3, power);
      pixel.b = mix(pixel.b, average, power);
      pixel.a = pixel.a*color.a;
      return pixel;
    }
    
    ]])
    
    
  shader.terrain = love.graphics.newShader([[
    

    vec4 effect( vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords )
    {
      vec4 pixel = Texel(texture, texture_coords )*color; //This is the current pixel color
      vec4 farPixel = Texel(texture, texture_coords/16 )*color;
      
      pixel = pixel + (farPixel - pixel)*0.3;
      
      return pixel;
    }
    
    ]])
    
  shader.timeBubble = love.graphics.newShader([[
    
    extern vec2 playerPos;
    extern vec2 cameraPos;
    extern float bubbleRadius;
    extern float zoom;
    extern float time;
    extern float roundAmount;
    
    vec4 effect( vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords )
    {
      
      float sinOffset = sin((( screen_coords.y - cameraPos.y )/ zoom )*0.25 + time*4)*zoom;
      vec4 pixel; 
      
      if (!(
      
      pow(floor((playerPos.x - screen_coords.x)/roundAmount + 0.5)*roundAmount - sinOffset, 2) + pow(floor((playerPos.y - screen_coords.y)/roundAmount + 0.5)*roundAmount, 2) 
      
      <= pow(bubbleRadius*zoom,2)
      
      
      ))
      {
        pixel = Texel(texture, vec2(texture_coords.x + sinOffset/1000, texture_coords.y) )*color; //This is the current pixel color
      
        number average = (pixel.r*color.r+pixel.g*color.g+pixel.b*color.b)/3.0;
      /*  pixel.r = (pixel.r + (average - pixel.r)*0.7f)*0.88f;
        pixel.g = (pixel.g + (average - pixel.g)*0.7f)*0.94f;
        pixel.b = (pixel.b + (average - pixel.b)*0.7f);*/
        
        pixel.r = pow(pixel.r, 1.3);
        pixel.g = pow(pixel.g, 1.2);
        pixel.b = pow(pixel.b, 1.1);
      }
      else
      {
        pixel = Texel(texture, texture_coords)*color; //This is the current pixel color
      }
      
      return pixel;
    }
    
    ]])
    
shader.setColour = love.graphics.newShader([[
    
    extern vec3 setColour;
    
    vec4 effect( vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords )
    {
    
      vec4 pixel = Texel(texture, texture_coords)*color; //This is the current pixel color
      number value = (pixel.r + pixel.g + pixel.b)/3.0;
      
      return vec4(setColour*value, pixel.a);
    }
    ]])
    
return shader