local programVersion = "Demo 0.1.10"

demoMode = true
startPaused = demoMode

paused = false

settings = {}
settings.cameraShake = 1

settings.particles = 1



if demoMode then
  programVersion = "Public ".. programVersion
end


local analyseFramerate = false
local framerateResults = {}
local framerateObjectCount = {}

--local screenCanvas


builderEnabled = true -- Lock down the engine if needed

currentFrame = 0
clickedThisFrameChecker = {}
clickedThisFrameChecker.primary = false
clickedThisFrameChecker.secondary = false
clickedThisFrame = {}
clickedThisFrame.primary = false
clickedThisFrame.secondary = false

tester = 0
scene = {}
scene.lightHarshness = 0.1

textureScale = 5


window = {}
mouse = {}
mouse.x = 0
mouse.y = 0
mouse.worldX = 0
mouse.worldY = 0
mouse.isDown = {}
mouse.lastMoved = 0 
controls = {}


fireflyTable = {}

--Table for which order things should be drawn in
local tableNumber = 0
require "util/miscFunctions"
require "lib/loadingAnalyser"

local sortTable = 
{
  "camera",
  "textEditor",
  "dialogueBox",
  "logicProp",
  "directionParticles",
  "grass",
  "firefly",
  "sword",
  "fireball",
  "sparks",
  "flowLevelPolygon",
  "levelPolygon",
  "raindrops",
  "player",
  "rainEmitter",
  "particleGen",
  "tempAnim",
  "slidingDoor",
  "door",
  "physicsCube",
  "flow",
  "interactable",
  "button",
  "talkative",
  "brazier",
  "prop",
  "directionStone",
  "backgroundPolygon",
  "cloudSheet",
  "other"
}

sortTable = swapKeyValues(sortTable)

local targetFps = 60



function love.load()  
  nextEntityID = 0
  debugMode = false
  screenshotMode = false
  builderMode = false

  drawTable = {}
  --table to help find named entities
  nameIndex = require "nameIndex"
  
  
  
  
  
  
  math.randomseed(os.time())
  
  
  Object = require "lib/classic"
  lume = require "lib/lume"
  utf8 = require "utf8"
  require "lib/blueGraphics"
  require "lib/vector"
  require "lib/line"
  require "lib/pixpol"
  require "load/events"
  require "point"
  require "textEditor"
  require "attributeEditor"
  require "entities/entity"
  require "entities/timeBubble"
  require "entities/prop"
  require "entities/builder"
  require "entities/polygon"
  require "entities/physicsCube"
  require "entities/levelPolygon"
  require "entities/cloudSheet"
  require "entities/camera"
  require "entities/grass"
  require "entities/particleGenerator"
  require "entities/firefly"
  require "entities/tempAnim"
  require "entities/projectile"
  require "entities/fireball"
  require "entities/sparks"
  require "entities/raindrops"
  
  require "entities/enemies/flow"
  
  require "entities/special props/logicProp"
  
  require "entities/interactables/interactable"
  require "entities/interactables/special interactables"
  require "entities/interactables/talkative"
  require "entities/interactables/character"
  require "entities/interactables/door"
  require "entities/interactables/slidingDoor"
  require "entities/interactables/brazier"
  require "entities/interactables/directionStone"
  
  require "entities/special props/rainEmitter"
  
  require "load/dialogue"
  require "load/level"
  require "load/pauseMenu"
  require "util/consoleOutput"
  require "util/polygonTest"
  require "entities/player"
  
  require "miscPhysics"
  require "animations"
  require "animationPlayer"
  
  
  
  --Load textures
  love.graphics.setDefaultFilter("nearest", "nearest", 4) 
  
  texture = {}
  
  texture.metatable = {}
  setmetatable(texture, texture.metatable)
  function texture.metatable.__index (table, key)
    return texture.placeholder
  end
  
  animation = {}
  
  animation.metatable = {}
  setmetatable(animation, animation.metatable)
  function animation.metatable.__index (table, key)
    --Don't have a placeholder animation yet
    return animation.drea_running
  end
  
  
  texture.keypress       = love.graphics.newImage("textures/misc/keypress.png")
  
  texture.cloudSheet     = love.graphics.newImage("textures/environment/cloud sheet.png")
  texture.stone          = love.graphics.newImage("textures/environment/stone.png")
  texture.wood           = love.graphics.newImage("textures/environment/wood.png")
  texture.bricks         = love.graphics.newImage("textures/environment/bricks.png")
  texture.vertical_wood  = love.graphics.newImage("textures/environment/vertical wood.png")
  texture.skyStone       = love.graphics.newImage("textures/environment/sky stone.png")
  texture.grass = {}
  texture.grass[1]       = love.graphics.newImage("textures/environment/grass/grass1.png")
  texture.grass[2]       = love.graphics.newImage("textures/environment/grass/grass2.png")
  texture.grass[3]       = love.graphics.newImage("textures/environment/grass/grass3.png")
  texture.grassFlower = {}
  texture.grassFlower[1] = love.graphics.newImage("textures/environment/grass/grass1flower.png")
  texture.grassFlower[2] = love.graphics.newImage("textures/environment/grass/grass2flower.png")
  texture.grassFlower[3] = love.graphics.newImage("textures/environment/grass/grass3flower.png")
  texture.stoneBox       = love.graphics.newImage("textures/objects/stoneBox.png")
  
  texture.logic_prop           = love.graphics.newImage("textures/icons/icon_logic.png")
  
  texture.grappling_hook   = love.graphics.newImage("textures/objects/grapplingHook.png")
  texture.backpack         = love.graphics.newImage("textures/objects/backpack.png")
  texture.rope_coil        = love.graphics.newImage("textures/objects/rope coil.png")
  texture.red_rope         = love.graphics.newImage("textures/environment/red rope.png")
  texture.stove            = textureSet("objects/stove")
  texture.kettle           = textureSet("objects/kettle")
  texture.barrel           = textureSet("objects/barrel")
  texture.table            = textureSet("objects/table")
  texture.chair            = textureSet("objects/chair")
  texture.pot              = textureSet("objects/pot")
  texture.light_wall       = textureSet("objects/light wall")
  texture.sign             = textureSet("objects/sign")
  texture.bookshelf        = textureSet("objects/bookshelf")
  texture.mug              = textureSet("objects/mug")
  texture.big_plant        = textureSet("objects/big plant")
  texture.blossom_painting = textureSet("objects/blossom painting")
  texture.brazier          = love.graphics.newImage("textures/objects/brazier.png")
  texture.brazier_lit      = love.graphics.newImage("textures/objects/brazier lit/brazier lit13.png")
  texture.direction_stone        = love.graphics.newImage("textures/objects/direction stone.png")
  texture.direction_stone_active = love.graphics.newImage("textures/objects/direction stone active.png")
  
  texture.icon_stone           = love.graphics.newImage("textures/icons/icon_stone.png")
  texture.icon_backgroundStone = love.graphics.newImage("textures/icons/icon_backgroundStone.png")
  texture.icon_wood            = love.graphics.newImage("textures/icons/icon_wood.png")
  texture.icon_backgroundWood  = love.graphics.newImage("textures/icons/icon_backgroundWood.png")
  texture.icon_prop            = love.graphics.newImage("textures/icons/icon_prop.png")
  texture.icon_talkative       = love.graphics.newImage("textures/icons/icon_talkative.png")
  texture.icon_rain            = love.graphics.newImage("textures/icons/icon_rain.png")

  
  
  texture.doors = {}
  texture.doors.metatable = {}
  setmetatable(texture.doors, texture.doors.metatable)
  function texture.doors.metatable.__index (table, key)
     return texture.doors.wooden
  end
  
  
  texture.doors.wooden = doorTextureSet("wooden", -18)
  
  texture.volt_mural_1 = love.graphics.newImage("textures/objects/murals/volt mural 1.png")
  texture.volt_mural_2 = love.graphics.newImage("textures/objects/murals/volt mural 2.png")
  texture.volt_mural_3 = love.graphics.newImage("textures/objects/murals/volt mural 3.png")
  texture.volt_mural_4 = love.graphics.newImage("textures/objects/murals/volt mural 4.png")
  
  
  texture.test_grom   = love.graphics.newImage("textures/test/grom_boop.png")
  texture.placeholder = love.graphics.newImage("textures/placeholder.png")
  
  texture.drea_idle     = textureSet("drea/Idle", 15.5)
  texture.drea_tucked   = textureSet("drea/Tucked")
  texture.drea_sleeping = textureSet("drea/Sleeping")
  texture.drea_falling  = textureSet("drea/Falling")
  texture.drea_sliding  = textureSet("drea/Sliding/Sliding", 16.5)
  
  texture.firefly      = love.graphics.newImage("textures/misc/Firefly.png")
  texture.firefly_body = love.graphics.newImage("textures/misc/Firefly Body.png")
  
  texture.magic_decay        = love.graphics.newImage("textures/effects/magic decay.png")
  texture.fireball           = love.graphics.newImage("textures/objects/fireball.png")
  texture.spark              = love.graphics.newImage("textures/effects/spark.png")
  texture.raindrop           = love.graphics.newImage("textures/effects/raindrop.png")
  texture.direction_particle = love.graphics.newImage("textures/effects/direction particle.png")
    
  animation.drea = {}
  animation.drea_running  = Animation("drea/Running",  13, 11, 0)
  animation.drea_flapping = Animation("drea/Flapping", 12, 4, -6)
  animation.drea_gliding  = Animation("drea/Gliding",  5, 4, 2)
  
  animation.effect = {}
  animation.effect_magicRipple = Animation("effects/magic ripple", 18)
  
  animation.kettle_steam = Animation("objects/kettle steam", 10)
  
  
  animation.brazier_lit = Animation("objects/brazier lit", 13, 0, 0)
  
  
  texture.voltehre_idle = textureSet("voltehre/idle")
  texture.cordelia_idle = love.graphics.newImage("textures/cordelia/idle.png")
  texture.curator_idle = love.graphics.newImage("textures/curator/idle.png")
  
  
  sound = {}
  sound.music = {}
  sound.effect = {}
  sound.music.hopesAndDreams = love.audio.newSource("sounds/Our Hopes And Dreams Looped.wav", "stream")
  sound.music.hopesAndDreams:setLooping(true)
  sound.music.ethereal = love.audio.newSource("sounds/60 Second Ethereal.mp3", "stream")
  
  sound.effect.thud = love.audio.newSource("sounds/Fast Banging on Dry Wall 1.mp3", "static")
  sound.effect.wingBeat = love.audio.newSource("sounds/Woosh C4-1.mp3", "static")
  sound.effect.wingBeat:setVolume(0.5)
  
  
  sound.effect.ui = {}
  sound.effect.ui.select   =  love.audio.newSource("sounds/UI Select.mp3", "static")
  sound.effect.ui.positive =  love.audio.newSource("sounds/UI Positive.mp3", "static")
  sound.effect.ui.positive:setVolume(0.25)
  sound.effect.ui.negative =  love.audio.newSource("sounds/UI Negative.mp3", "static")
  
  sound.effect.firefly_collect = love.audio.newSource("sounds/Firefly Collect.mp3", "static")
	sound.effect.firefly_collect:setVolume(0.8)
  sound.effect.bounce_parry = love.audio.newSource("sounds/Bounce Parry.mp3", "static")
  sound.effect.bounce_parry:setVolume(0.5)
  
  sound.effect.stasis = love.audio.newSource("sounds/Stasis Splash.mp3", "static")
  sound.effect.stasis:setVolume(0.7)
  sound.effect.boost = love.audio.newSource("sounds/Boost.mp3", "static")
  sound.effect.direction_stone_activate = love.audio.newSource("sounds/Direction Stone Activate.mp3", "static")
  sound.effect.direction_stone_activate:setVolume(0.6)
  
  
  shader = require "load/shaders"
  
  
  if demoMode then
    sound.music.hopesAndDreams:play()
  end
  musicVolume = -0.1
  sound.music.hopesAndDreams:setVolume(0)
  
  smallTextureScale = 40/(texture.drea_idle.main:getHeight() - 2)
  
  -- Basic physics code based on love2D tutorial
  -- the height of a meter in our world will be 64px
  
  love.physics.setMeter(64)
  -- create a world for the bodies to exist in with horizontal gravity
  -- of 0 and vertical gravity of 9.81
  gravity = 9.81*64
  
  currentLevel = Level("nil")
  
  pauseMenu = Menu("Pause Menu", 
    {
      MenuElement("Resume"), 
      MenuElement("Settings"), 
      MenuElement("Quit", love.event.quit)
    })
  
  world = love.physics.newWorld(0, gravity, true)
  world:setCallbacks( beginContact, endContact, preSolve, postSolve )
  player1 = Player(-10460, -2650)
  builder = Builder(0,0)
  camera = Camera(0, 0, 1, 1)
  createSparkGroup()
  createRaindropGroup()
  createDirectionParticleGroup()
  
  
  love.window.setMode(0,0)
  love.keyboard.setKeyRepeat( true)
  
  
  loadingScreen()
  
  
  love.window.setVSync(0)
  love.mouse.setVisible(false)
  window.width = love.graphics.getWidth()
  window.height = love.graphics.getHeight()
  --tmpCanvas = love.graphics.newCanvas(window.width, window.height)
  screenCanvas = love.graphics.newCanvas(window.width, window.height)
  
  
  font = love.graphics.newImageFont("textures/font.png", 
  "ABCDEFGHIJKLMNOPQRSTUVWXYZ"..
  "abcdefghijklmnopqrstuvwxyz"..
  "0123456789().,-=+\"':;/?! *_|@", 1)

  love.graphics.setFont(font)
  
  textEditor = TextEditor()
  dialogueBox = DialogueBox(window.width/2, window.height*6.8/8, nil)
 
  for i = 0, 3 do
    
    love.math.setRandomSeed(i)
    local layerDepth = i*5 + 1.0
    table.insert(drawTable, CloudSheet(love.math.random(-100000, 100000), love.math.random(-100000, 100000), layerDepth))    
    
  end
  
  
  
  currentLevel:loadBakedLevel("baked saved level")
  
  
  love.mouse.setPosition(window.width/2, window.height/2)
  
end


function love.update(dt)
  
  
    
    local numberOfUpdatesPerFrame = 1
    dt = dt / numberOfUpdatesPerFrame
    if (startPaused and demoMode) then
      manageControls()
      if controls.interact.isDown then
        startPaused = false
        love.mouse.setPosition(window.width/2, window.height/2)
        sound.music.hopesAndDreams:seek(0.67)
        
      end
    else
      for i = 1, numberOfUpdatesPerFrame do
        currentFrame = currentFrame + 1
        local j = 0
        manageControls()
        
        clickedThisFrame.primary = false
        clickedThisFrame.secondary = false
        
        if love.mouse.isDown(1) then
          if not clickedThisFrameChecker.primary then
            clickedThisFrame.primary = true
          end
          clickedThisFrameChecker.primary = true
        else
          clickedThisFrameChecker.primary = false
        end
        
        if love.mouse.isDown(2) then
          if not clickedThisFrameChecker.secondary then
            clickedThisFrame.secondary = true
          end
          clickedThisFrameChecker.secondary = true
        else
          clickedThisFrameChecker.secondary = false
        end
        
        
        
        
        
        if musicVolume < 1 then
          musicVolume = musicVolume + dt*0.5
          sound.music.hopesAndDreams:setVolume(musicVolume)
        elseif musicVolume > 1 then
          musicVolume = 1
          sound.music.hopesAndDreams:setVolume(1)
        end
        mouse.lastX = mouse.x
        mouse.lastY = mouse.y 
        
        
        
        local joysticks = {}--love.joystick.getJoysticks()
        
        
        
        if #joysticks > 0 then
          local distance = window.height/2
          joystick = joysticks[1]
          mouse.x = joystick:getGamepadAxis("leftx")*distance
          mouse.y = joystick:getGamepadAxis("lefty")*distance
        
        else
          mouse.x = love.mouse.getX() - window.width/2
          mouse.y = love.mouse.getY() - window.height/2
        end
        

        mouse.worldX = mouse.x/camera.zoom - camera.x
        mouse.worldY = mouse.y/camera.zoom - camera.y

        
        
        
        if not (mouse.x == mouse.lastX and mouse.y == mouse.lastY) then
          mouse.lastMoved = love.timer.getTime()
        end
        
        if paused then
          
          
          if controls.menuUp.thisFrame then
            pauseMenu:up()
          elseif controls.menuDown.thisFrame then
            pauseMenu:down()
          elseif controls.menuEnter.thisFrame then
            pauseMenu:enter()
          end
          
          
        else
        if builderMode then
          
          if not textEditor.active then
            builder:update(dt)
          
            camera.x = builder.x * -1
            camera.y = builder.y * -1
            
            for i, object in ipairs(drawTable) do
              if object.updateBuilder ~= nil then
                object:updateBuilder(dt)
              end
            end
            flushDrawTable()
          end
        else
        
          slowMotion = 1
          dt = dt * slowMotion
          
          
          if 1/dt < 60 then
            dt = 1/60
            --print("limited "..love.timer.getTime())
          end
          
          if not dialogueBox.pauseWorld then
            player1:clearCollisionTable()
            --player1.body:setFixedRotation(false)
          else
            --player1.body:setFixedRotation(true)
          end
          
          
          world:update(dt)
          camera:update(dt)
          
          builder.x = camera.x * -1
          builder.y = camera.y * -1
          
          for i, object in ipairs(drawTable) do
            if object.isInteractable then
              --cout(object.type, currentLevel.stage)
            end
            
            if object.update ~= nil then
              if not dialogueBox.pauseWorld or object == dialogueBox then
                
                -- Only update if current stage. If not, set as inactive
                if object.activeStage == currentLevel.stage or object.activeStage == nil then
                  
                  --Update and set active
                  if object.body and (not object.body:isDestroyed()) then
                    object.body:setActive(true)
                  end
                  object:update(dt)
                  
                  
                  
                else
                  
                  if object.body and (not object.body:isDestroyed()) then
                    object.body:setActive(false)
                  end
                end
                
              end
            end
          end
          flushDrawTable()
          
          
        end
      end
    end
  end
end



function flushDrawTable()
  --Remove unwanted objects if tagged with 'remove'
  for i, object in ipairs(drawTable) do
    if object.remove then
      
      if object.attributes and object.attributes.name and object.attributes.name ~= "" then
        nameIndex:removeEntity(object)
      end
      
      table.remove(drawTable, i)
    end
  end
end


function love.draw()

  texture.drea_idle.main:setFilter("nearest", "nearest", anisotropicFiltering)
  texture.drea_idle.shadow:setFilter("nearest", "nearest", anisotropicFiltering)
  
  if startPaused and demoMode then
    
    love.graphics.setBackgroundColor(0,0,0)
    local scale = 2.5 * (window.height/1080) --Adjust for smaller screens
    local safeZone = 100
    love.graphics.setColor(1,1,1)
    love.graphics.printf(
      [[Hello there! This is a demo of my game, thanks for checking it out!
      
      If you want help me develop this further, sending me a video of you playing would be a great help!
      
      I also have a YouTube channel at www.youtube.com/@bobble68, which I *sometimes* upload update videos to.
      
      - Bobble
      
      
      Press ]]..controls.interact.name.." to start", safeZone, safeZone, (love.graphics.getWidth()-2*safeZone)/scale, "center", 0, scale, scale)
  else
    
    --[[do 
      local t = math.min(player1.magic/3, 1)
      
      love.graphics.setBackgroundColor(
      math.lerp(135/255, 98/255, t), 
      math.lerp(190/255, 149/255, t),
      math.lerp(235/255, 191/255, t))
    end]]
    
    
    love.graphics.setColor(1,1,1,1)
    
    table.sort(drawTable, 
    function(k1, k2)
      --True if k2 should be in front of k1
      

      local k1z = k1.z or 1
      local k2z = k2.z or 1
      
      --Objects with a lower Z will always be returned first
      if k1z ~= k2z then
        return k1z > k2z
      end
      
      local k1Type = k1.sortType or k1.type or "other"
      local k2Type = k2.sortType or k2.type or "other"
      

      if sortTable[k1Type] == nil then
        error(k1Type .." does not exist in the sort table!")
      end
      if sortTable[k2Type] == nil then
        error(k2Type .." does not exist in the sort table!")
      end
      
      if k1Type == "levelPolygon" and k2Type == "levelPolygon" then
        if k1 == builder.selectedPolygon or k1 == builder.activePolygon then
          return false
        end
        if k2 == builder.selectedPolygon or k2 == builder.activePolygon then
          return true
        end
        

        if k1.attributes.isBackground then
          return true
        end
        if k2.attributes.isBackground then
          return false
        end

    end
    
    
    if k1Type == k2Type then
      if k1.attributes then
        if k1.attributes.priorityDraw then
          return false
        elseif k2.attributes.priorityDraw then
          return true
        end
      end
    end


      return sortTable[k1Type] > sortTable[k2Type]
    end
  )
    
    if player1.timeBubble then
      love.graphics.setCanvas(screenCanvas)
      love.graphics.clear(132/255, 173/255, 1)
    else
      love.graphics.setBackgroundColor(132/255, 173/255, 1)
    end
    for i, v in ipairs(drawTable) do
      
      
      if v.activeStage == currentLevel.stage or v.activeStage == nil then
        
        local renderStart = nil
        if analyseFramerate then
          renderStart = love.timer.getTime()
        end
        if v.type ~= "levelPolygon" or (v.type == "levelPolygon" and v.finalised) then
          v:draw()
        elseif builderMode then
          v:draw()
        end
        if analyseFramerate then
          
          if framerateResults[v.type] then
            framerateResults[v.type] = framerateResults[v.type] + love.timer.getTime() - renderStart
            framerateObjectCount[v.type] = framerateObjectCount[v.type] + 1 
          else
            framerateResults[v.type] = love.timer.getTime() - renderStart
            framerateObjectCount[v.type] = 1
          end
          

        end
      end
    end
    if player1.timeBubble then
      love.graphics.setCanvas()
      love.graphics.setColor(1,1,1)
      
      local playerScreenX, playerScreenY = blue.standardTransform(player1.body:getX(), player1.body:getY())
      local cameraX, cameraY = blue.standardTransform(0, 0)
      
      shader.timeBubble:send("playerPos", {playerScreenX, playerScreenY})
      shader.timeBubble:send("cameraPos", {cameraX, cameraY})
      shader.timeBubble:send("bubbleRadius", player1.timeBubble)
      shader.timeBubble:send("zoom", camera.zoom)
      shader.timeBubble:send("time", love.timer.getTime())
      shader.timeBubble:send("roundAmount", smallTextureScale*camera.zoom)
      
      love.graphics.setShader(shader.timeBubble)
      love.graphics.draw(screenCanvas)    
      love.graphics.setShader()
    end
    --[[blurCanvas(glowCanvas, 5)
    
    
    love.graphics.setColor(1,1,1)
    love.graphics.setBlendMode("alpha", "premultiplied")
    love.graphics.draw(glowCanvas, 0, 0)
    love.graphics.setBlendMode("alpha")]]
    
    printInstructions()
    textEditor:draw()
    if not builderMode then
      love.graphics.setColor(1,1,1)
    else
      builder:draw()
      love.graphics.setColor(0.5,0.9,1,0.3)
    end
    
    if not screenshotMode then
    --Draw mouse
    
    
      if love.timer.getTime() - mouse.lastMoved < 3 then
        love.graphics.circle("fill", mouse.x + window.width/2, mouse.y + window.height/2, 5)
        love.graphics.setColor(0,0,0)
        love.graphics.circle("line", mouse.x + window.width/2, mouse.y + window.height/2, 5)
      end
      love.graphics.setColor(1,0, 0)
      --blue.graphics:pixellatedDraw(texture.test.grom, 0, 0, 0, 1)
      
      local pointTable = {-100, -100, 100, 0, 150, 100}


      
      --blue.graphics.circle("fill", mouse.x - camera.x, mouse.y - camera.y, 5)
      love.graphics.setColor(0,0,0)
      --blue.graphics.circle("line", mouse.x - camera.x, mouse.y - camera.y, 5)
      
      --blue.graphics.circle("fill", 0,0, 100)
      
      
          
      if builderMode then
        blue.graphics.printShadow(
          "fps:  "     ..love.timer.getFPS( )..
          "\nx:  "     ..(math.floor(mouse.worldX*100)/100)..
          "\ny:  "     ..(math.floor(mouse.worldY*100)/100)..
          "\nZoom:  "  ..(math.floor(camera.zoom*100)/100)..
          "\nTester:  "..tostring(tester)..
          "\nMode:  "  ..string.upper(builder.mode), 10,10)

      elseif debugMode then
        blue.graphics.printShadow(
          "fps:  "   .. love.timer.getFPS( )..
          "\nv:  "   ..math.floor(vector.getMagnitude(velocityVector)/64) .. "m/s"..
          "\nZoom:  "..(math.floor(camera.zoom*1000)/1000)..
          "\nMemory useage:  "..math.floor(collectgarbage("count")).. " kb"..
          "\nTester:  "..tostring(tester), 10,10)
      
      end
      console:draw()
      love.graphics.setColor(0,0,0)
      
      blue.graphics.printShadow(programVersion, 10, math.floor(window.height)-20) 
      love.graphics.setColor(1,1,1)
    end
    
    
    
    
    --glowCanvas:renderTo(love.graphics.clear)
  end
  
  
  --pauseMenu:draw()
  
end

function printInstructions()
  love.graphics.setColor(1,1,1)
  local xOffset, yOffset = blue.standardTransform(-10400, -2750)
  
  love.graphics.print(
      [[
      Use left and right
      or A and D to run
      ]], xOffset, yOffset, 0, camera.zoom)
      
  xOffset, yOffset = blue.standardTransform(-9000, -2850)
  love.graphics.print(
      [[
      Hold shift to
      tuck into a ball
      ]], xOffset, yOffset, 0, camera.zoom)
      
  xOffset, yOffset = blue.standardTransform(-7550, -2350)
  love.graphics.print(
      [[
      Point with your mouse
      and press space to fly
      ]], xOffset, yOffset, 0, camera.zoom)
end


function love.keypressed(key, scancode, isrepeat )
  
  
  if not startPaused then
    if (not textEditor.active) then
      if key == "escape" and not builderMode then
        pauseMenu:show()
      elseif key == "escape" then
        builder:hideMenu()
      elseif key ==  "\\" then
        debugMode = toggle(debugMode, true, false )
      
        
      elseif key == "m" then
        player1.magic = player1.magic + 1
        
      elseif key == "f1" then
        screenshotMode = not screenshotMode
        
      elseif builderEnabled then
        
        if key == "tab" then
          if builderMode and not (builder.selectedObject and builder:isModeSelectable()) then
            builder:toggleMenu()
          end
        
        elseif not builder.showBuilderMenu then
        
          
          if key ==  "b" then
            screenshotMode = false
            builderMode = not builderMode
            builder:hideMenu()
            
          elseif key == "f5" and not isrepeat then
            currentLevel:saveBakedLevel("baked saved level")
            cout("Saved baked!")
            
          elseif key == "f2" then
            currentLevel:deload()
            cout("Cleared world!")
            
          elseif key == "f8" and not isrepeat then
            currentLevel:loadBakedLevel("baked saved level")
            builder.selectedObject = nil
            cout("Loaded baked level")
          elseif key == "/"  then
         
            cout("Select Stage")
            textEditor.callback = function()
              currentLevel:setStage(textEditor.value)
            end
            
            textEditor:show(currentLevel.stage)
            textEditor.value = currentLevel.stage
            
          end
            
          --[[elseif key == "t" then
            if builderMode and (builder.mode == "prop" or builder.mode == "talkative") then
              textEditor.callback = function()
                builder.defaultPropTexture = textEditor.value
              end
              textEditor:show(builder.defaultPropTexture)
              textEditor.value = builder.defaultPropTexture
            end
          end]]
        end
        
      end
      
    
  
    else
      if key == "backspace" then
        textEditor.value = string.sub(textEditor.value, 1, textEditor.cursorPos-1).. string.sub(textEditor.value, textEditor.cursorPos+1, #textEditor.value)
        textEditor.cursorPos = math.max(textEditor.cursorPos - 1, 0)
      elseif key == "delete" then
        textEditor.value = string.sub(textEditor.value, 1, textEditor.cursorPos).. string.sub(textEditor.value, textEditor.cursorPos+2, #textEditor.value)
      elseif key == "return" then
        textEditor:finish()
      elseif key == "escape" then
        textEditor:close()
      elseif key == "right" then
        textEditor.cursorPos = math.min(textEditor.cursorPos + 1, #textEditor.value)
      elseif key == "left" then
        textEditor.cursorPos = math.max(textEditor.cursorPos - 1, 0)
      elseif key == "v" and (not clipboard) and love.keyboard.isScancodeDown("lctrl") then
        local clipboard = love.system.getClipboardText( )
        textEditor.value = string.sub(textEditor.value, 1, textEditor.cursorPos)..clipboard.. string.sub(textEditor.value, textEditor.cursorPos+1, #textEditor.value)
        textEditor.cursorPos = textEditor.cursorPos + #clipboard
        
      end
    end
  end
end



--[[function manageMouseImputs()
  
  mouse.lastMoved = love.timer.getTime()
  
  if love.mouse.isDown(1) then
    if mouse.thisFrame.primaryReleased then
      mouse.thisFrame.primary = true
      mouse.thisFrame.primaryReleased = false
    end
  end
  if love.mouse.isDown(1) then
    mouse.thisFrame.primaryReleased = false
  else
    mouse.thisFrame.primaryReleased = true
  end

    
end]]


function textureSet(filename, xOffset, yOffset)
  local textureSet = {}
  
  textureSet.ox = xOffset or 0
  textureSet.oy = yOffset or 0

  
  
  textureSet.main = love.graphics.newImage("textures/"..filename..".png")
  if file_exists("textures/"..filename.." Shadow.png") then
    textureSet.shadow = love.graphics.newImage("textures/"..filename.." Shadow.png")
  end
  if file_exists("textures/"..filename.." Glow.png") then
    textureSet.glow = love.graphics.newImage("textures/"..filename.." Glow.png")
  end
  return textureSet
end


function doorTextureSet(name, xOpenOffset)
  local textureSet = {}
  
  textureSet.ox = xOpenOffset or 0
    
  textureSet.open   = love.graphics.newImage("textures/objects/doors/"..name.." door open.png")
  textureSet.closed = love.graphics.newImage("textures/objects/doors/"..name.." door closed.png")

  return textureSet
end



function loadingScreen()
  
  local scale = love.graphics.getHeight()/texture.drea_tucked.main:getHeight()*0.1
  local r = love.timer.getTime()*4
  love.graphics.clear()
  love.graphics.setColor(1,1,1)
  love.graphics.draw(texture.drea_tucked.main,
    love.graphics.getWidth() - texture.drea_tucked.main:getWidth()*scale,
    love.graphics.getHeight() - texture.drea_tucked.main:getHeight()*scale,
    r, scale, scale,
    texture.drea_tucked.main:getWidth()/2,
    texture.drea_tucked.main:getHeight()/2
  )
  
  love.graphics.setColor(0, 0, 0, 0.15)
  
  love.graphics.draw(texture.drea_tucked.shadow,
    love.graphics.getWidth() - texture.drea_tucked.main:getWidth()*scale,
    love.graphics.getHeight() - texture.drea_tucked.main:getHeight()*scale,
    r, scale, scale,
    texture.drea_tucked.main:getWidth()/2,
    texture.drea_tucked.main:getHeight()/2
    )
  
  love.graphics.present()
  
end




function manageControls()
  
  --[[
  Here you can remap the controls. The first argument refers to the action and should be left unchanged, the second 
  is the condition that needs to be met to activate the control, and the third argument is what it will be referred to 
  as in game.
  ]]
  
  controlController("interact",      love.keyboard.isDown("f"),              "F")
  controlController("flap",          love.keyboard.isScancodeDown("space") or  (joystick and joystick:isDown(1)),  "space")
  controlController("tuck",          love.keyboard.isScancodeDown("lshift") or love.keyboard.isScancodeDown("rshift"), "left shift")
  controlController("stasis",        love.keyboard.isDown("e"),      "E")
  controlController("timeBubble",    love.keyboard.isDown("1"),      "1")
  controlController("boost",         love.keyboard.isScancodeDown("lctrl"),  "left ctrl")
  controlController("lockDirection", love.keyboard.isScancodeDown("lalt"),   "left alt")
  
  controlController("menuLeft",      love.keyboard.isDown("q"),   "Q")
  controlController("menuRight",     love.keyboard.isDown("e"),   "E")
  controlController("menuUp",        love.keyboard.isDown("w") or love.keyboard.isDown("up"),   "W")
  controlController("menuDown",      love.keyboard.isDown("s") or love.keyboard.isDown("down"), "S")
  controlController("menuEnter",     love.keyboard.isDown("return"),  "Enter")
  
end


function controlController(name, isDown, buttonName)
  if not controls[name] then
    controls[name] = {}
  end
  controls[name].name = buttonName
  if isDown then
    controls[name].thisFrame = not controls[name].isDown
    controls[name].upThisFrame = false
    controls[name].isDown = true
  else
    controls[name].upThisFrame = controls[name].isDown
    controls[name].isDown = false
    controls[name].thisFrame = false
  end
end



function love.quit()
  console:dumpOutput()
  
  
  
  
  
  if analyseFramerate then
    local normalise = false
    local totalRenderTime = 0
    
    for i, v in pairs(framerateResults) do
      if normalise then
        totalRenderTime = totalRenderTime + v/framerateObjectCount[i]
      else
        totalRenderTime = totalRenderTime + v
      end
      
    end
    
    if not normalise then
      for i, v in pairs(framerateResults) do
        print(i, math.floor(v/totalRenderTime*100000)/1000 .."%" )
      end
    else
      for i, v in pairs(framerateResults) do
        print(i, math.floor(((v/framerateObjectCount[i])/totalRenderTime*100000)/1000) .."%" )
      end
    end
  end
end


--[[function blurCanvas(self, amount)
  
  love.graphics.setColor(1,1,1)
  shader.glow:send("texture_size", {window.width, window.height})
  love.graphics.setBlendMode("alpha", "premultiplied")
  
  for i = 1, amount do
    tmpCanvas:renderTo(
      function()
        love.graphics.clear()
        shader.glow:send("horizontal", false)
        love.graphics.setShader(shader.glow)
        love.graphics.draw(self, 0, 0)
        love.graphics.setShader()
      end
    )
    
    self:renderTo(
      function()
        love.graphics.clear()
        shader.glow:send("horizontal", true)
        love.graphics.setShader(shader.glow)
        love.graphics.draw(tmpCanvas, 0, 0)
        love.graphics.setShader()
      end
    )
  end
  love.graphics.setBlendMode("alpha")

end]]


